use crate::container::stop_containers_by_ids;
use bollard::container::{AttachContainerOptions, LogOutput};
use bollard::Docker;
use futures_util::future::join_all;
use futures_util::StreamExt;
use std::collections::HashMap;
use std::iter::Map;

pub(crate) async fn monitor_build_outputs(
    docker: &Docker,
    ids: HashMap<String, String>,
) -> anyhow::Result<()> {
    let mut handles = vec![];
    for id in ids.values() {
        let docker = docker.clone();
        let id = id.clone();
        handles.push(tokio::spawn(async move {
            let r = monitor_logs(&docker, id.clone()).await;
            println!("Container {} exited with {:?}", id, r)
        }));
    }

    tokio::select! {
        _ = tokio::signal::ctrl_c() => {
            println!("Ctrl-c received, stopping containers");
            stop_containers_by_ids(docker, ids.values().map(|v| v.to_string()).collect()).await?;
        }
        _ = join_all(handles) => {
            println!("All containers exited");
        }
    }
    Ok(())
}

async fn monitor_logs(docker: &Docker, id: String) -> anyhow::Result<()> {
    let mut attach_results = docker
        .attach_container(
            &id,
            Some(AttachContainerOptions::<String> {
                stdout: Some(true),
                stderr: Some(true),
                stdin: Some(false),
                stream: Some(true),
                ..Default::default()
            }),
        )
        .await?;

    while let Some(log_result) = attach_results.output.next().await {
        match log_result {
            Ok(chunk) => match chunk {
                LogOutput::StdIn { .. } => unreachable!(),
                LogOutput::Console { .. } => unreachable!(),
                LogOutput::StdOut { message } => {
                    println!("{}", String::from_utf8_lossy(&message));
                }
                LogOutput::StdErr { message } => {
                    println!("{}", String::from_utf8_lossy(&message));
                }
            },
            Err(e) => println!("{}", e),
        }
    }
    Ok(())
}
